<?php

namespace Drupal\diboo_core\Plugin\ExtraField\Display;

use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\extra_field\Plugin\ExtraFieldDisplayFormattedBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Extra field plugin OpenChains.
 *
 * The bundles will be changed later in
 * diboo_core_extra_field_display_info_alter()
 * to display only on those that are rooms.
 *
 * @ExtraFieldDisplay(
 *   id = "diboo_core_open_chains",
 *   label = @Translation("Open Chains"),
 *   bundles = {
 *     "node.*",
 *   }
 * )
 */
final class OpenChains extends ExtraFieldDisplayFormattedBase implements ContainerFactoryPluginInterface {

  use StringTranslationTrait;

  /**
   * Constructs a OpenChains object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\Core\Datetime\DateFormatterInterface $dateFormatter
   *   The date formatter.
   * @param \Drupal\Core\Session\AccountProxyInterface $currentUser
   *   The current user.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    private EntityTypeManagerInterface $entityTypeManager,
    private DateFormatterInterface $dateFormatter,
    private AccountProxyInterface $currentUser
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): self {
    return new static(
      $configuration, $plugin_id, $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('date.formatter'),
      $container->get('current_user'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getLabel() : TranslatableMarkup {
    return $this->t('Open chains');
  }

  /**
   * {@inheritdoc}
   */
  public function getLabelDisplay(): string {
    return 'above';
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(ContentEntityInterface $room) : array {
    /** @var \Drupal\diboo_core\Entity\Room $room */
    $openChains = $room->getOpenChains();
    $viewBuilder = $this->entityTypeManager->getViewBuilder('node');
    $elements = [
      '#cache' => [
        'contexts' => ['user'],
        // @todo Define a better tag with bundles that are chains like node_list:BUNDLE.
        'tags' => ['node_list'],
      ],
    ];
    if (!$this->currentUser->hasPermission('add chain links to chains')) {
      $elements[] = [
        '#markup' => $this->t('<a href="@register">Register</a> to participate', [
          '@register' => Url::fromRoute('user.register')->toString(),
        ]),
      ];
      return $elements;
    }

    if (!$openChains) {
      $elements[] = [
        '#markup' => $this->t('No chains open'),
      ];
      return $elements;
    }

    /** @var \Drupal\diboo_core\Entity\Chain $chain */
    foreach ($openChains as $chain) {
      $render = $viewBuilder->view($chain, 'diboo_open_chain');
      $render['#theme'] = 'node__diboo_open_chain';

      $chainLinks = $chain->get('diboo_chain_links')->referencedEntities();
      $lastChainLink = end($chainLinks);
      $lastChainLinkType = $lastChainLink->get('type')->first()->get('entity')->getValue();

      $render['chain_links_count'] = [
        '#markup' => count($chainLinks),
      ];
      $render['last_chain_link_type'] = [
        '#markup' => strtolower($lastChainLinkType->label()),
      ];
      $render['last_author'] = [
        '#markup' => $lastChainLink->getOwner()->getDisplayName(),
      ];
      $render['last_chain_link_date'] = [
        '#markup' => $this->dateFormatter->format($lastChainLink->getCreatedTime(), 'short'),
      ];

      // @todo check for anonymous as current contributor.
      $currentAuthorId = $chain->get('diboo_current_contributors')->first()?->get('target_id')->getValue();
      if ($currentAuthorId !== NULL) {
        /** @var \Drupal\user\Entity\User $currentContributor */
        $currentContributor = $this->entityTypeManager->getStorage('user')->load($currentAuthorId);
        $render['current_contributor'] = ['#markup' => $currentContributor->getDisplayName()];
      }

      $availableChainLinkTypes = $this->getAvailableChainLinkTypes($lastChainLink->bundle());
      $links = [];
      foreach ($availableChainLinkTypes as $chainLinkType) {
        $url = Url::fromRoute(
          'diboo_core.add_chain_link',
          [
            'node_type' => $chainLinkType->id(),
            'chain' => $chain->id(),
          ],
          [
            'query' => ['destination' => $room->toUrl()->toString()],
          ]
        );
        // @todo Retrieve reasons for the access denied and display them.
        if (!$url->access()) {
          continue;
        }
        $isImage = $chainLinkType->getThirdPartySetting('diboo_core', 'chain_link_type') === 'image';
        $links['add_chain_link_' . $chainLinkType->id()] = [
          'title' => $isImage ? $this->t('Draw') : $this->t('Describe'),
          'url' => $url,
        ];
      }

      $elements[] = [
        '#type' => 'details',
        '#title' => $this->t('Chain started by @author on @date', [
          '@author' => $chain->getOwner()->getDisplayName(),
          '@date' => $this->dateFormatter->format($chain->getCreatedTime(), 'short'),
        ]),
        'details' => $render,
      ];
      $elements[] = [
        '#type' => 'operations',
        '#links' => $links,
      ];
    }
    return $elements;
  }

  /**
   * Gets a list of available chain link types.
   *
   * @param string $chainLinkTypeToExclude
   *   Remove from the set.
   *
   * @return \Drupal\node\NodeTypeInterface[]
   *   Node types.
   */
  protected function getAvailableChainLinkTypes(string $chainLinkTypeToExclude = ''): array {
    $nodeTypes = $this->entityTypeManager->getStorage('node_type')->loadMultiple();
    return array_filter($nodeTypes, function ($nodeType) use ($chainLinkTypeToExclude) {
      if ($nodeType->id() === $chainLinkTypeToExclude) {
        return FALSE;
      }
      $dibooFunction = $nodeType->getThirdPartySetting('diboo_core', 'diboo_function');
      if ($dibooFunction === 'link') {
        return TRUE;
      }
      return FALSE;
    });
  }

}
