<?php

namespace Drupal\diboo_core\Plugin\Field\FieldFormatter;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\Plugin\DataType\EntityAdapter;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldFormatter\EntityReferenceFormatterBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Routing\RouteProviderInterface;
use Drupal\Core\Url;
use Drupal\Core\Utility\Token;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Format an entity reference as one of its links templates.
 *
 * This formatter is generic enough to live in its own module.
 *
 * @FieldFormatter(
 *   id = "entity_link",
 *   label = @Translation("Entity link"),
 *   field_types = {
 *     "entity_reference"
 *   }
 * )
 */
final class EntityLinkFormatter extends EntityReferenceFormatterBase {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The token service.
   *
   * @var \Drupal\Core\Utility\Token
   */
  protected $token;

  /**
   * The token service.
   *
   * @var \Drupal\Core\Routing\RouteProviderInterface
   */
  protected $routeProvider;

  /**
   * The token service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * Constructs an EntityReferenceEntityFormatter instance.
   *
   * @param string $plugin_id
   *   The plugin_id for the formatter.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The definition of the field to which the formatter is associated.
   * @param array $settings
   *   The formatter settings.
   * @param string $label
   *   The formatter label display setting.
   * @param string $view_mode
   *   The view mode.
   * @param array $third_party_settings
   *   Any third party settings.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Utility\Token $token
   *   The token service.
   * @param \Drupal\Core\Routing\RouteProviderInterface $routeProvider
   *   The route provider service.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $moduleHandler
   *   The route provider service.
   */
  public function __construct(
    $plugin_id,
    $plugin_definition,
    FieldDefinitionInterface $field_definition,
    array $settings,
    $label,
    $view_mode,
    array $third_party_settings,
    EntityTypeManagerInterface $entity_type_manager,
    Token $token,
    RouteProviderInterface $routeProvider,
    ModuleHandlerInterface $moduleHandler
  ) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $label, $view_mode, $third_party_settings);
    $this->entityTypeManager = $entity_type_manager;
    $this->token = $token;
    $this->routeProvider = $routeProvider;
    $this->moduleHandler = $moduleHandler;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): self {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['label'],
      $configuration['view_mode'],
      $configuration['third_party_settings'],
      $container->get('entity_type.manager'),
      $container->get('token'),
      $container->get('router.route_provider'),
      $container->get('module_handler'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings(): array {
    return [
      'link_template' => '',
      'link_text' => '',
      'route_parameter_first' => '',
      'route_parameter_context_first' => 'displayed_entity',
      'route_parameter_second' => '',
      'route_parameter_context_second' => 'displayed_entity',
      'route_parameter_third' => '',
      'route_parameter_context_third' => 'displayed_entity',
      'destination' => '',
      'destination_context' => 'displayed_entity',
    ] + parent::defaultSettings();
  }

  /**
   * Extract the entity type from the current chosen template.
   *
   * The format is ENTITY_TYPE:LINK_TEMPLATE_ID.
   */
  protected function getCurrentEntityType(): string {
    $linkTemplate = $this->getSetting('link_template');
    if (!$linkTemplate) {
      return $this->getFieldSetting('target_type');
    }
    [$entityType] = explode(':', $linkTemplate);
    return $entityType;
  }

  /**
   * Extract the entity type from the current chosen template.
   *
   * The format is ENTITY_TYPE:LINK_TEMPLATE_ID.
   */
  protected function getCurrentTemplateId(): string {
    $linkTemplate = $this->getSetting('link_template');
    if (!$linkTemplate) {
      return 'canonical';
    }
    [, $templateId] = explode(':', $linkTemplate);
    return $templateId;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state): array {
    $form = parent::settingsForm($form, $form_state);

    $options = [];
    foreach ($this->entityTypeManager->getDefinitions() as $entityTypeId => $entityType) {
      $linkTemplates = $entityType->getLinkTemplates();
      // Ignore entity types with no link templates.
      if (!$linkTemplates) {
        continue;
      }
      array_walk($linkTemplates, function (&$path, $linkTemplateId) {
        $path = $linkTemplateId . ': ' . $path;
      });
      foreach ($linkTemplates as $linkTemplateId => $path) {
        $options[$entityTypeId . ':' . $linkTemplateId] = $entityType->getLabel() . ' - ' . $path;
      }
    }

    $form['link_template'] = [
      '#title' => $this->t('Link template'),
      '#type' => 'select',
      '#required' => TRUE,
      '#options' => $options,
      '#default_value' => $this->getSetting('link_template') ?: ($this->getCurrentEntityType() . ':canonical'),
    ];

    $form['link_text'] = [
      '#title' => $this->t('Link text'),
      '#type' => 'textfield',
      '#required' => TRUE,
      '#default_value' => $this->getSetting('link_text'),
    ];

    $form['route_parameter_first'] = [
      '#title' => $this->t('First route parameter'),
      '#type' => 'textfield',
      '#description' => $this->t('Fill values for the parameters as they appear on the path'),
      '#placeholder' => '[node:nid]',
      '#default_value' => $this->getSetting('route_parameter_first'),
    ];

    $form['route_parameter_context_first'] = [
      '#title' => $this->t('First route parameter context'),
      '#type' => 'select',
      '#options' => [
        'referenced_entity' => $this->t('Use referenced entity'),
        'displayed_entity' => $this->t('Use entity being displayed'),
      ],
      '#description' => $this->t('For token replacements, chose the the entity that will be provided.'),
      '#default_value' => $this->getSetting('route_parameter_context_first'),
    ];

    $form['route_parameter_second'] = [
      '#title' => $this->t('Second route parameter'),
      '#type' => 'textfield',
      '#default_value' => $this->getSetting('route_parameter_second'),
    ];

    $form['route_parameter_context_second'] = [
      '#title' => $this->t('Second route parameter context'),
      '#type' => 'select',
      '#options' => [
        'referenced_entity' => $this->t('Use referenced entity'),
        'displayed_entity' => $this->t('Use entity being displayed'),
      ],
      '#description' => $this->t('For token replacements, chose the the entity that will be provided.'),
      '#default_value' => $this->getSetting('route_parameter_context_second'),
    ];

    $form['route_parameter_third'] = [
      '#title' => $this->t('Third route parameter'),
      '#type' => 'textfield',
      '#default_value' => $this->getSetting('route_parameter_third'),
    ];

    $form['route_parameter_context_third'] = [
      '#title' => $this->t('Third route parameter context'),
      '#type' => 'select',
      '#options' => [
        'referenced_entity' => $this->t('Use referenced entity'),
        'displayed_entity' => $this->t('Use entity being displayed'),
      ],
      '#description' => $this->t('For token replacements, chose the the entity that will be provided.'),
      '#default_value' => $this->getSetting('route_parameter_context_third'),
    ];

    $form['destination'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Destination'),
      '#description' => $this->t('For redirecting back to a URL a ?destination=URL will be added.'),
      '#placeholder' => '[node:url:path]',
      '#default_value' => $this->getSetting('destination'),
    ];
    $form['destination_context'] = [
      '#title' => $this->t('Destinatination context'),
      '#type' => 'select',
      '#options' => [
        'referenced_entity' => $this->t('Use referenced entity'),
        'displayed_entity' => $this->t('Use entity being displayed'),
      ],
      '#description' => $this->t('For token replacements, chose the the entity that will be provided.'),
      '#default_value' => $this->getSetting('destination_context'),
    ];
    if ($this->moduleHandler->moduleExists('token')) {
      $form['token_help'] = [
        '#theme' => 'token_tree_link',
        '#token_types' => [
          $this->fieldDefinition->getTargetEntityTypeId(),
          $this->getFieldSetting('target_type'),
        ],
      ];
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary(): array {
    $summary = [];
    $summary[] = $this->t('Link template: @link_template', ['@link_template' => $this->getSetting('link_template')]);
    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function viewElements(FieldItemListInterface $items, $langcode): array {
    $elements = [];

    foreach ($this->getEntitiesToView($items, $langcode) as $delta => $entity) {
      if ($entity->isNew()) {
        continue;
      }
      $entityType = $this->entityTypeManager->getDefinition($this->getCurrentEntityType());
      $path = $entityType->getLinkTemplate($this->getCurrentTemplateId());
      $routes = $this->routeProvider->getRoutesByPattern($path)->all();
      if (!$routes) {
        continue;
      }
      $routeName = array_key_first($routes);
      $route = reset($routes);

      $params = [];
      if (!empty($route->getOptions()['parameters'])) {
        $parameterPosition = ['first', 'second', 'third'];
        $currentPosition = 0;
        foreach ($route->getOptions()['parameters'] as $parameterName => $parameterConfig) {
          $parameterValue = $this->getSetting('route_parameter_' . $parameterPosition[$currentPosition]);
          $parameterContext = $this->getSetting('route_parameter_context_' . $parameterPosition[$currentPosition]);
          $currentPosition++;
          if (!$parameterValue) {
            continue;
          }
          // Guess if there is tokens.
          if (strpos(haystack: $parameterValue, needle: '[') === FALSE) {
            $params[$parameterName] = $parameterValue;
          }
          // For now only support entities for replacements.
          if (empty($parameterConfig['type']) || !str_starts_with($parameterConfig['type'], 'entity:')) {
            continue;
          }
          // Reduce, for example, entity:node to just "node".
          $parameterEntityType = str_replace(search: 'entity:', replace: '', subject: $parameterConfig['type']);
          if ($parameterContext === 'displayed_entity' && $items->getParent() instanceof EntityAdapter) {
            $contextEntity = $items->getParent()->getEntity();
          }
          else {
            $contextEntity = $entity;
          }
          if ($parameterEntityType !== $contextEntity->getEntityTypeId()) {
            continue;
          }
          $params[$parameterName] = $this->token->replace($parameterValue, [
            $parameterEntityType => $contextEntity,
          ]);
        }
      }
      $options = [];
      if ($destination = $this->getSetting('destination')) {
        $contextEntity = $this->getSetting('destination_context') === 'displayed_entity' ?
          $items->getParent()->getEntity() :
          $entity;
        $options['query'] = [
          'destination' => $this->token->replace($destination, [$contextEntity->getEntityTypeId() => $contextEntity]),
        ];
      }

      $url = Url::fromRoute($routeName, $params, $options);
      if (!$url->access()) {
        continue;
      }

      $elements[$delta] = [
        '#type' => 'link',
        '#title' => $this->token->replace($this->getSetting('link_text'), [$entity->getEntityTypeId() => $entity]),
        '#url' => $url,
      ];
      $elements[$delta]['#cache']['tags'] = $entity->getCacheTags();
    }

    return $elements;
  }

}
